# -*- coding: utf-8 -*-
"""
Created on Fri Mar 18 12:21:50 2022

@author: acferreira
"""

# Import OGR - 
from osgeo import ogr

# Import GDAL
from osgeo import gdal, gdal_array

# Open the dataset from the file
shapefile_fn="G:/My Drive/QUARENTENA_COVID19/DETECAO_REMOTA_MULTIESPETRAL/PRATICAS_DRM/DATA/reference_samples.shp"
shape_ds = ogr.Open(shapefile_fn)

 # Make sure the dataset exists -- it would be None if we couldn't open it
if not shape_ds:
    print('Error: could not open dataset')
    
### What is the name of the layer?
layer = shape_ds.GetLayerByIndex(0)
print('The layer is named: {n}\n'.format(n=layer.GetName()))

### How many features are in the layer?
feature_count = layer.GetFeatureCount()
print('Layer has {n} features\n'.format(n=feature_count))    

### How many fields are in the shapefile, and what are their names?
# First we need to capture the layer definition
defn = layer.GetLayerDefn()

# How many fields
field_count = defn.GetFieldCount()
print('Layer has {n} fields'.format(n=field_count))

# What are their names?
print('Their names are: ')
for i in range(field_count):
    field_defn = defn.GetFieldDefn(i)
    print('\t{name} - {datatype}'.format(name=field_defn.GetName(),
                                         datatype=field_defn.GetTypeName()))
    

# First we will open our raster image, to understand how we will want to rasterize our vector
raster_fn= "G:/My Drive/QUARENTENA_COVID19/DETECAO_REMOTA_MULTIESPETRAL/PRATICAS_DRM/DATA/subset_0_of_collocate.tif"
raster_ds = gdal.Open(raster_fn, gdal.GA_ReadOnly)

# Fetch number of rows and columns
ncol = raster_ds.RasterXSize
nrow = raster_ds.RasterYSize
count= raster_ds.RasterCount
print("Number of bands of the image:",count)



# Fetch projection and extent
proj = raster_ds.GetProjectionRef()
ext = raster_ds.GetGeoTransform()

raster_ds = None

# Create the raster dataset
memory_driver = gdal.GetDriverByName('GTiff')
out_raster_ds = memory_driver.Create('G:/My Drive/QUARENTENA_COVID19/DETECAO_REMOTA_MULTIESPETRAL/PRATICAS_DRM/DATA/reference_samples.gtif', ncol, nrow, 1, gdal.GDT_Int32)

# Set the ROI image's projection and extent to our input raster's projection and extent
out_raster_ds.SetProjection(proj)
out_raster_ds.SetGeoTransform(ext)

# Fill our output band with the 0 blank, no class label, value
b = out_raster_ds.GetRasterBand(1)
b.Fill(0)


# Rasterize the shapefile layer to our new dataset
status = gdal.RasterizeLayer(out_raster_ds,  # output to our new dataset
                             [1],  # output to our new dataset's first band
                             layer,  # rasterize this layer
                             None, None,  # don't worry about transformations since we're in same projection
                             [0],  # burn value 0
                             ['ALL_TOUCHED=TRUE',  # rasterize all pixels touched by polygons
                              'ATTRIBUTE=code']  # put raster values according to the 'id' field values
                             )

# Close dataset
out_raster_ds = None

if status != 0:
    print("I don't think it worked...")
else:
    print("Success")

# Import NumPy for some statistics
import numpy as np

roi_ds = gdal.Open('G:/My Drive/QUARENTENA_COVID19/DETECAO_REMOTA_MULTIESPETRAL/PRATICAS_DRM/DATA/reference_samples.gtif', gdal.GA_ReadOnly)

roi = roi_ds.GetRasterBand(1).ReadAsArray()

# How many pixels are in each class?
classes = np.unique(roi)
# Iterate over all class labels in the ROI image, printing out some information
for c in classes:
    print('Class {c} contains {n} pixels'.format(c=c,
                                                 n=(roi == c).sum()))

# The values in roi correspond to the numeric code of each class to be used for classification
print('The values in roi are:',np.unique(roi))



# To visualize both data

raster_ds = gdal.Open(raster_fn, gdal.GA_ReadOnly)
img = np.zeros((raster_ds.RasterYSize, raster_ds.RasterXSize, raster_ds.RasterCount),
               gdal_array.GDALTypeCodeToNumericTypeCode(raster_ds.GetRasterBand(1).DataType))

for b in range(img.shape[2]):
    img[:, :, b] = raster_ds.GetRasterBand(b + 1).ReadAsArray()
 
print(img.shape)
print(img)
    
roi = roi_ds.GetRasterBand(1).ReadAsArray().astype(np.uint8)

import matplotlib.pyplot as plt
plt.close('all')

# Display them
plt.figure(1)
plt.subplot(121)
plt.imshow(img[:, :, 7], cmap=plt.cm.Greys_r)
plt.title('NIR')

plt.subplot(122)
plt.imshow(roi, cmap=plt.cm.Spectral)
plt.title('ROI Training Data')

plt.show()

# Find how many non-zero entries we have -- i.e. how many training data samples?
n_samples = (roi > 0).sum()
print('We have {n} samples'.format(n=n_samples))

# What are our classification labels? Exclude class 0 that corresponds to the pixels that do not overlay with polygons
labels = np.unique(roi[roi > 0])
print('The training data include {n} classes: {classes}'.format(n=labels.size, 
                                                                classes=labels))
# We will need a "X" matrix containing our features, and a "y" array containing our labels
#     These will have n_samples rows
#     In other languages we would need to allocate these and them loop to fill them, but NumPy can be faster

X = img[roi > 0, :]  
y = roi[roi > 0]

print('Our X matrix is sized: {sz}'.format(sz=X.shape))
print('Our y array is sized: {sz}'.format(sz=y.shape))

print (X)
print(X.shape)
print (y)

###############################################################################################################
# To especify only certain variables from your X array use np.r_, in the case of an array with 26 variables 
# if you want to select variables B2 (column 1), B3 (column 2), B8 (column 7), B11 (column 10), B12 (column 11) - numpy arrays start at ZERO! 
# https://www.pythonprogramming.in/how-to-select-specific-columns-in-numpy-array.html

# number of bands for each acquisition date
# bands=13 # half of 26 in this case


# xdata=[]
# for i in range(int(len(X[0])/bands)):
#     list=X[:, np.r_[(i*bands)+1:(i*bands)+3, (i*bands)+7, (i*bands)+10:(i*bands)+12]]
#     xdata.append(list)

# X_array=np.array(xdata)
# print(X_array, X_array.shape)


# Y=X_array.transpose(1,0,2)
# print(Y, Y.shape)

# X=Y.reshape(Y.shape[0],-1)
# print(X, X.shape)

# #### Updates the new number of variables for FEATURE IMPORTANCE!
# # get the number of columns
# count=len(X[0])


###############################################################################################################


# Training the RANDOM FOREST

from sklearn.ensemble import RandomForestClassifier
from sklearn.model_selection import train_test_split

# Split the training data into train and validation
X_train, X_test, y_train, y_test = train_test_split (X, y, test_size=0.40, random_state=42)

print('Our X_train matrix is sized: {sz}'.format(sz=X_train.shape))
print('Our y_train array is sized: {sz}'.format(sz=y_train.shape))



# Initialize our model with 500 trees
rf = RandomForestClassifier(n_estimators=500, oob_score=True,n_jobs=-1)

# Fit our model to training data
rf = rf.fit(X_train, y_train)
print(rf)

# Check out the "Out-of-Bag" (OOB) prediction score

print('Our OOB prediction of accuracy is: {oob}%'.format(oob=rf.oob_score_ * 100)) 

# Check out the feature importance scores to get an idea of which spectral bands are more relevant

bands= np.arange(1,count+1,1)
print(bands)

for b, imp in zip(bands, rf.feature_importances_):
    print('Band {b} importance: {imp}'.format(b=b, imp=imp))
    
### https://mljar.com/blog/feature-importance-in-random-forest/    
    
plt.figure(2)

plt.barh (bands, rf.feature_importances_)

# Sort the features by name
#bands_name=['B1_M', 'B2_M', 'B3_M', 'B4_M', 'B5_M', 'B6_M','B7_M', 'B8_M', 'B8A_M','B9_M', 'B11_M', 'B12_M','simple_ratio_M','B1_S', 'B2_S', 'B3_S', 'B4_S', 'B5_S', 'B6_M','B7_S', 'B8_S', 'B8A_S','B9_S', 'B11_S', 'B12_S','simple_ratio_S']
#sorted_idx = rf.feature_importances_.argsort()
#plt.barh (bands_name[sorted_idx], rf.feature_importances_[sorted_idx])

plt.xlabel("Random Forest Feature Importance")
plt.show()
print (bands.shape)
print(rf.feature_importances_.shape)


## Classify the entire image after the Random Forest classifier fit

## Take our full image and reshape into long 2d array (nrow * ncol, nband) for classification
new_shape = (img.shape[0] * img.shape[1], img.shape[2])
bd=img.shape[2]
img_as_array = img[:, :, :bd].reshape(new_shape)
print('Reshaped from {o} to {n}'.format(o=img.shape,
                                        n=img_as_array.shape))

###############################################################################################################
# To especify only certain variables from your X array use np.r_, in the case of an array with 26 variables 
# if you want to select variables B2 (column 1), B3 (column 2), B8 (column 7), B11 (column 10), B12 (column 11) - numpy arrays start at ZERO! 
# https://www.pythonprogramming.in/how-to-select-specific-columns-in-numpy-array.html

# number of bands for each acquisition date
# bands=13 # half of 26 in this case


# xdata_img=[]
# for i in range(int(len(img_as_array[0])/bands)):
#     list=img_as_array[:, np.r_[(i*bands)+1:(i*bands)+3, (i*bands)+7, (i*bands)+10:(i*bands)+12]]
#     xdata_img.append(list)

# img_array=np.array(xdata_img)
# print(img_array, img_array.shape)


# T=img_array.transpose(1,0,2)
# print(T, T.shape)

# img_as_array=T.reshape(T.shape[0],-1)
# print(img_as_array, img_as_array.shape)



####################################################################################################################

# Now predict for each pixel in the entire image
class_prediction = rf.predict(img_as_array)

print(class_prediction.shape)

# Reshape our classification map
class_prediction = class_prediction.reshape(img[:, :, 0].shape)

print(class_prediction)
print(class_prediction.shape)


#Evaluating the Algorithm
#Confusion matrix, precision recall, and F1 values

from sklearn.metrics import classification_report, confusion_matrix, accuracy_score, cohen_kappa_score
from sklearn.metrics import balanced_accuracy_score

# Predict for the validation pixels
y_pred=rf.predict(X_test)

print(confusion_matrix(y_test,y_pred))
print(classification_report(y_test,y_pred))
print(accuracy_score(y_test, y_pred))
print(balanced_accuracy_score(y_test, y_pred))
print(cohen_kappa_score(y_test, y_pred))

import sklearn
print('sklearn: {}'.format(sklearn.__version__))

# Visualize the result of the classification

# First setup a 5-4-3 composite
#def color_stretch(image, index, minmax=(0, 10000)):
#    colors = image[:, :, index].astype(np.float64)
#
#    max_val = minmax[1]
#    min_val = minmax[0]
#
#    # Enforce maximum and minimum values
#    colors[colors[:, :, :] > max_val] = max_val
#    colors[colors[:, :, :] < min_val] = min_val
#
#    for b in range(colors.shape[2]):
#        colors[:, :, b] = colors[:, :, b] * 1 / (max_val - min_val)
#        
#    return colors
#    
#img543 = color_stretch(img, [4, 3, 2], (0, 8000))
#
## See https://github.com/matplotlib/matplotlib/issues/844/
#n = class_prediction.max()
#
#
## Next setup a colormap for our map
#colors = dict((
#    (0, (0, 0, 0, 255)),  # Nodata
#    (3, (0, 150, 0, 255)),  # Forest
#    (6, (0, 0, 255, 255)),  # Water
#    (4, (0, 255, 0, 255)),  # Herbaceous
#    (5, (160, 82, 45, 255)),  # Bare soil
#    (2, (100, 82, 0, 255)),  # Agriculture
#    (1, (255, 0, 0, 255))  # Artificial
#))
#### Put 0 - 255 as float 0 - 1
#for k in colors:
#    v = colors[k]
#    _v = [_v / 255.0 for _v in v]
#    colors[k] = _v
#    
#index_colors = [colors[key] if key in colors else 
#                (255, 255, 255, 0) for key in range(1, n + 1)]
#cmap = plt.matplotlib.colors.ListedColormap(index_colors, 'Classification', n)

# Now show the classmap next to the image
plt.figure (3)
#plt.subplot(121)
#plt.imshow(img543)
#
#plt.subplot(122)
#plt.imshow(class_prediction, cmap=cmap, interpolation='none')

plt.imshow(class_prediction,cmap=plt.cm.Spectral)
plt.show()

# Exporting the classification map to GeoTIFF format

# Create the raster dataset
memory_driver = gdal.GetDriverByName('GTiff')
classif = memory_driver.Create('G:/My Drive/QUARENTENA_COVID19/DETECAO_REMOTA_MULTIESPETRAL/PRATICAS_DRM/DATA/classification.gtif', ncol, nrow, 1, gdal.GDT_Int32)

# Set the classification image's projection and extent to our input raster's projection and extent
# https://gis.stackexchange.com/questions/150300/how-to-place-an-numpy-array-into-geotiff-image-using-python-gdal
classif.SetProjection(proj)
classif.SetGeoTransform(ext)
outband=classif.GetRasterBand(1)
outband.WriteArray(class_prediction)

# Close the dataset
classif = None
